/*** This file is part of the logit program. (C) 1999/2000 by S. Krahmer
 *** This program can be used under the terms of the GPL.
 *** There is absolutely NO WARRANTY for this program.
 *** You use this at your own risk! 
 *** 
 *** The logit package contains code for strong crypto (RC4).
 *** Be sure that you don't get any conflicts with the law of your
 *** country. I'm not responsible for any damage or trouble you'll
 *** maybe get due to using this program.
 ***
 *** S. Krahmer 	<krahmer@cs.uni-potsdam.de>
 ***/
#include "portable.h"
#include "logit.h"
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

/* main() forks the process, opens a communication channel
 * for the applications who wish to log sth., and passes their
 * messages to the printLOG() function.
 */
int main(void)
{
        int sockfd = 0, afd = 0, kfd = 0, current_fd = 0;
	FILE *fd[LFCOUNT] = {NULL};
        char logbuf[MAXLOG + 11] = {0};
        int pid = 0, fd_index = 0;
        char *s = NULL;    
	fd_set rset;
	struct timeval tv;

   	umask(037);
        
   	if ((clilog = fopen(MYCLILOG, "a+")) == NULL) {
           	fprintf(stderr, "Fatal: Cannot open internal logfile (%s)\n", MYCLILOG);
                exit(errno);
        }
                
        /* create the channel */
        sockfd = createSocket(LOGPATH);
        
        /* fork it */
#ifndef DEBUG
        if ((pid = fork()) < 0) {
           	perror("fork:");
                exit(errno);
        } else if (pid > 0) {
           	printf(LOGIT_MSG);
                printDate(clilog); fprintf(clilog, "[logit] started.\n");
                exit(0);
        }
#endif
        /* fd is an array here! */
        openLogfiles(fd);
        /* if found any loghosts, start a second process wich takes
         * his messages via a pipe
         */
        if (forkcount) {
           	signal(SIGCHLD, reaper);
           	startx();
        }

	if ((kfd = open(KLOG, O_RDONLY)) < 0) {
		perror("open");
		exit(errno);
	}
        while (1) {
		FD_ZERO(&rset);
		FD_SET(sockfd, &rset);
		FD_SET(kfd, &rset);
		
		if (select((kfd < sockfd ? sockfd+1:kfd+1), &rset, NULL, NULL, NULL) < 0) 
			continue;
		
	/*
#ifndef DEBUG
	        if ((pid = fork()) > 0) 
			continue;
		else if (pid < 0) {
			perror("fork");
			exit(errno);
		} else {
#endif	*/
                   	if (FD_ISSET(sockfd, &rset)) {
                           	if ((afd = accept(sockfd, NULL, 0)) < 0) {
                                   	perror("accept");
                                        exit(errno);
                                }
				
				/* we don't want someone to block after accept
				 */
                                current_fd = afd;
				FD_ZERO(&rset);
				FD_SET(afd, &rset);
				tv.tv_sec = 3;
				select(afd + 1, &rset, NULL, NULL, &tv);
				if (!FD_ISSET(afd, &rset)) {
					close(afd);
					continue;
				}
                        } else if (FD_ISSET(kfd, &rset))
                           	current_fd = kfd;
			else
				continue;
                                	      
	    		/* and parse...*/
			read(current_fd, logbuf, MAXLOG);
            		sscanf(logbuf, "<%d>", &fd_index);
            		s = strchr(logbuf, '>');
            		if (!s) 
                           	s = logbuf;
    			fd_index = fd_index >> 3;
			if (fd_index < 0 || fd_index > LFCOUNT - 1)
                   		fd_index = LOG_USER >> 3; 
				
			/* if read raw from /proc/kmsg we need to add the date/time
			 */
            		if (current_fd == kfd)
				printDate(fd[fd_index]);
				
			printLOG(fd[fd_index], s+1);
            		memset(logbuf, 0, MAXLOG + 11); 

			/* only close if fd comes from accept */
			if (current_fd != kfd)
                       		close(current_fd);
#ifndef DEBUG
		/*	exit(0);
		}*/
#endif
        }
	/* if we went here, sth. must be wrong */
	perror("???");
        return errno;
}

/* openLogfiles() does'nt only open files.
 * It parses the configfile, and looks whether a specified location
 * is a file or a loghost (url). If the location is a host,
 * it creates a pipe instead of a normal file. This pipe is used by main()
 * as a normal descriptor in fd[] array. But on the other end of the pipe
 * listens startx() to get main's() messages and pass it over the network.
 *
 */
int 
openLogfiles(FILE *fd[])
{
   	int i = 0;
        struct hostent *hos;
        int pdes[2] = {0};
        
        parseConfigfile(CONFIG);
        /* Now LF contains the file- and hostnames */
        for (i = 0; i < LFCOUNT; i++) {
           	h_errno = 0;
                
		/* check if we find a url */
                if (!strchr(LF[i].file, '/')) {
			hos = gethostbyname(LF[i].file);
            		if (!h_errno) {
                   		pipe(pdes);
#ifdef DEBUG
                    		printf("creating pipes: %d %d\n", pdes[0], pdes[1]);
#endif
                    		/* open a pipe if a host was found */
                    		pipes[forkcount].fd[0] = pdes[0];
                    		pipes[forkcount].fd[1] = pdes[1];    
				
                    		/* save URL */            
                    		strncpy(pipes[forkcount].url, LF[i].file, 256);
                   		/* and make it a stream from the pipedes */
                    		if ((fd[i] = fdopen(pipes[forkcount].fd[1], "w")) == NULL) {
					perror("fdopen");
					exit(errno);
				}
#ifdef DEBUG
                    		printf("added %d to fd[i] with destination %s\n", pipes[forkcount].fd[1],pipes[forkcount].url);
#endif
                    		/* count the found hosts */
                    		forkcount++;
            		} else
                   	/* the default behaivior: open a logfile */
                   	fd[i] = openLOG(LF[i].file);
		} else
		/* if a '/' in the location it's also a file */
			fd[i] = openLOG(LF[i].file);
        }
        return 0;
}
        
/* This starts a process wich is connected via pipes to the
 * main flow. It does a select over all open pipes and handles
 * the messages that come in.
 */
int 
startx(void)
{
   	int pid = 0, i = 0, maxfd = 0;
        fd_set rdes;
        
        if ((pid = fork()) < 0) {
           	printDate(clilog); fprintf(clilog, "[logit] fork failed (errno = %d)\n", errno);
                return errno;
        /* else go back to main() */
        } else if (pid > 0)
           	return 0;
        
        while (1) {
           	maxfd = 0;
           	FD_ZERO(&rdes);
                /* switch on the read-fd flag for select on every pipe we have*/
                for (i = 0; i < LFCOUNT && pipes[i].url[0]; i++) {
#ifdef DEBUG
                   	printf("switch on fd %d for select()\n", pipes[i].fd[0]);
#endif
                        FD_SET(pipes[i].fd[0], &rdes);
                        /* calc first arg to select() */
                        if (maxfd < pipes[i].fd[0])
                           	maxfd = pipes[i].fd[0];
                }
                maxfd++;
           	select(maxfd, &rdes, NULL, NULL, NULL);
                for (i = 0; i < LFCOUNT && pipes[i].url[0]; i++) {
                   	if (FD_ISSET(pipes[i].fd[0], &rdes))
                           	cliProto(i);
                }
        }
}
