/*** This file is part of the logit program. It's under the GPL.
 *** You use this on your own risk. I'm not responsible for any
 *** damage or trouble you may be get due to using this program.
 *** Take a look at rfc.-1 for a deeper description for this proto.
 *** S. Krahmer 	<krahmer@cs.uni-potsdam.de>
 ***/
#include "portable.h"
#include "logit.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/select.h>
#include <string.h>
#include "rc4/rc4.h"
#include "md5/md5.h"

int 
getRandomstring(char *s, int len)
{ 
     int fd = 0, i = 0;
     
     
     if ((fd = open("/dev/random", O_RDONLY)) < 0) {
        	perror("open");
                exit(errno);
     }
     for (i = 0; i < len; i++) {
             if (read(fd, &s[i], 1) < 0) {
        		printDate(srvlog); fprintf(srvlog, "[logserv] can't read from /dev/random - exiting (errno=%d)\n", errno);
            		exit(errno);
             }
     }
     return 0;
}

/* The server-side of the protocol */
int
srvProto(FILE *fd)
{
   	FILE *log;
   	struct timeval tv;
        fd_set rset;
        char request[2*RLEN] = {0}, /* the request-random-string */
             rstring[RLEN] = {0}, 
             response[2*RLEN] = {0},  /* the MD5Sum, 16 Bytes */
             sessionkey[2*RLEN] = {0}, /* the request rc4 encrypted with MASTERKEY */
             message[2*(MAXIDENT+MAXLOG+100)] = {0};
        char *s;
        int i = 0, t = 0,
            dont_close = 0; /* look if we use same descriptor for internal and normal log */
        
        if ((log = fopen(SRVLOG, "a+")) == NULL) {
           	printDate(srvlog); 
                fprintf(srvlog, "[logserv] Can't open '%s' (errno = %d).\n", SRVLOG, errno);
                dup2(fileno(srvlog), t);        
                /* Ey, we run out of fd's, so mark that we use same des */
                if ((log = fdopen(t, "a+")) == NULL) {
                   	dont_close++;
                   	log = srvlog;
                }
        }                
        tv.tv_sec = 3;
        tv.tv_usec = 0;
        FD_ZERO(&rset);
        FD_SET(fileno(fd), &rset);
        
        /* Give nobody the chance to block us */
        select(fileno(fd) + 1, &rset, NULL, NULL, &tv);
        if (!FD_ISSET(fileno(fd), &rset)) {
           	printDate(log); fprintf(log, "[logserv] timeout.\n");
           	fclose(fd);
                return 0;
        }
        fgets(request, 10, fd);

        if (strcmp(request, "GREETS\n")) {
           	fprintf(fd, "Unknown request.\n");
                printDate(log); fprintf(log, "[logserv] Other side says: %s\n", request);
                if (!dont_close)
                   	fclose(log);
                return 0;
        }
        getRandomstring(rstring, RLEN - 1);
        /* put it into the right format */
        for (i = 0; i < RLEN - 1; i++) 
           	sprintf(request+2*i, "%02x", (unsigned char)rstring[i]);
        
        fprintf(fd, "%s\n", request);
        fflush(fd);
#ifdef DEBUG
        printf("server: my request is %s\n", request);
#endif
        /* this call will change 'request' to an my_encrypted string */
        calcResponse(request, response);
#ifdef DEBUG
        printf("server: expected response is %s\n", response);
#endif
        /* we take this string for the sessionkey */
        for (i = 0; i < RLEN - 1; i++)
           	sprintf(sessionkey+2*i, "%02x", (unsigned char)request[i]);
        
        tv.tv_sec = 3;
        tv.tv_usec = 0;
        FD_ZERO(&rset);
        FD_SET(fileno(fd), &rset);
        /* again: no DoS ! */
        select(fileno(fd) + 1, &rset, NULL, NULL, &tv);
        if (!FD_ISSET(fileno(fd), &rset)) {
           	fclose(fd);
                return 0;
        }
        fgets(request, 2*RLEN, fd);
        /* If he fails: */
        if (strncmp(response, request, strlen(response)) != 0) {
           	fclose(fd);
                printDate(srvlog);
                fprintf(srvlog, "[logserv] He cannot answer my question.\n");
                if (!dont_close)
                   	fclose(log);
                return 0;
        }
#ifdef DEBUG
        printf("sessionkey: %s response: %s\n", sessionkey, response);
#endif
        /* get the real message (my_encrypted) */
        fgets(message, 2*(MAXLOG+MAXIDENT+100), fd);
        
#ifdef DEBUG
        printf("got %s\n", message);
#endif
        message[strlen(message)-1] = '\0';
        s = (char*)Malloc(strlen(message)+1);
        my_decrypt(s, message, sessionkey);
#ifdef DEBUG
        printf("%s\n", s);
#endif
        fprintf(log, "%s", s);
        if (!dont_close)
           	fclose(log);
        fprintf(fd, "QUIT\n"); fflush(fd);
        return 0;
}
   	
int        
calcResponse(char* request, char response[2*RLEN])
{
   	rc4_key rc4key;
        struct MD5Context md5Sum;
        unsigned char t[RLEN] = {0};
        int i = 0;
 
#ifdef DEBUG
    	printf("the MASTERKEY is %s\n", MASTERKEY);       
#endif
        prepare_key((unsigned char*)MASTERKEY, strlen(MASTERKEY), &rc4key);
        rc4((unsigned char*)request, strlen(request), &rc4key);
        MD5Init(&md5Sum);
        MD5Update(&md5Sum, (unsigned char*)request, strlen(request));
        MD5Final(t, &md5Sum);	
        for (i = 0; i < RLEN - 1; i++) 
           	sprintf(response+2*i, "%02x", (unsigned char)t[i]);
        return 0;
}

/* my_encrypt the string s2 into s1 wich has our special %02x format
 */
int
my_encrypt(char *s1, char *s2, char *key)
{
   	rc4_key rc4key;
        int i = 0, s2len = 0;
        
        s2len = strlen(s2);
   	prepare_key((unsigned char*)key, strlen(key), &rc4key);
        rc4((unsigned char*)s2, strlen(s2), &rc4key);
        for (i = 0; i < s2len; i++)
           	sprintf(s1+2*i, "%02x", (unsigned char)s2[i]);
        return 0;
}

/* my_decrypt from s2 (remeber %02x) into s1 */
int 
my_decrypt(char *s1, char *s2, char *key)
{
   	rc4_key rc4key;
        int i = 0,
            hex = 0, s1len = 0;
        
        s1len = strlen(s2)/2;
        
        prepare_key((unsigned char*)key, strlen(key), &rc4key);
        for (i = 0; i < s1len; i++) {
                sscanf(s2+2*i, "%02x", &hex);
                s1[i] = (unsigned)hex;
        }
        rc4((unsigned char*)s1, s1len, &rc4key);
        return 0;
}

/* the client side implementation of our proto 
 * 'index' is the index of the array 'pipes' wich holds the descriprotrs
 * and the host-location.
 */
int 
cliProto(int index)
{
   	int i = 0;
   	char request[2*RLEN] = {0},	/* because its in %02x format */
             response[2*RLEN] = {0},
             message[MAXIDENT+MAXLOG+100] = {0},
             sessionkey[2*RLEN] = {0};
        char *s = NULL;     
        FILE *f;
        
        /* get da message from whom select was interrupted */
        read(pipes[index].fd[0], message, MAXIDENT+MAXLOG+100);
        
        /* open a stream to the server */
        if ((f = fdopen(makeconnection(pipes[index].url, LOGITPORT), "a+")) == NULL) {
           	printDate(clilog); 
                fprintf(clilog, "[logit] Can't open stream to server. (errno = %d) log it here:", errno);
                fprintf(clilog, "%s", message); fflush(clilog);
           	return -1;
        }
        /* the protocol begins: */
        fprintf(f, "GREETS\n"); fflush(f);
        
        /* get the request string from him */
        fgets(request, 2*RLEN, f);
        
        request[strlen(request)-1] = '\0'; /* skip newline */
#ifdef DEBUG
        printf("client: request is %s\n", request);
#endif
        calcResponse(request, response); 
#ifdef DEBUG
        printf("client: my response is %s\n", response);
#endif
        /* tell him our computed response */
        fprintf(f, "%s\n", response); fflush(f);
        
        /* and use rc4'ed request for sessionkey */
        for (i = 0; i < RLEN - 1; i++) 
           	sprintf(sessionkey+2*i, "%02x", (unsigned char)request[i]);
#ifdef DEBUG
        printf("sessionkey: %s\n", sessionkey);
        printf("message: %s\n", message);
#endif
        s = (char*)Malloc(2*strlen(message)+1);
        my_encrypt(s, message, sessionkey);               
#ifdef DEBUG
        printf("send %s\n", s);
#endif
        /* send the my_encrypted message */
        fprintf(f, "%s\n", s); fflush(f);
        
        /* wait for the "QUIT" */
        fgets(request, 10, f);
        fclose(f);
        return 0;
}

